
! ESA (C) 2000-2024 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
!+
! 
!    Package        : epauxcheck
!
!    Creation date  : 27-Jun-2024
!
!    Description    : Check PNAUX file for later use in CCD pn chain
!
!    Authors        :
!
!    
!		Michael Freyberg - MPE - mjf@mpe.mpg.de
!		
!
!-

module epauxcheck_mod

  use dal
  use param  
  use errorHandling

  implicit none 

  private 

  public epauxcheck

contains 

  !---------------------------------------------------------------------
  subroutine epauxcheck 
  !---------------------------------------------------------------------

  implicit none
  character(len=13), parameter :: CodeVersion = '0.7'

  character(len=1024)          :: fileName, auxfileName
  logical(kind=bool)           :: withpnauxname, withexplanation
  logical(kind=bool)           :: hasPNAUX1, hasPNAUX2
  integer(kind=Int32)          :: ntaux1, ntaux2
  integer(kind=Int32)          :: nffff, nover, nloss, n48828, n32400, ftc_max(0:3)
  integer(kind=Int32)          :: iquad, irow, it, ntmax
  character(len=40)            :: auxName
  character(len=22)            :: odf_icd
  character(len=5)             :: epn_submode(0:7)

  logical(kind=bool)           :: ldone = .false.
  integer(kind=Int32)          :: ical, ip(1:2), fd, ih, hh, hh1, hh2, hh3, anchop
  real(kind=8)                 :: ysu, y1, y2, ftfav(0:3), pp(1:4), ftav1_exp, ftav1_aux, freq_diff
  real(kind=8)                 :: ftd_frms, ftd_frac, ftd_corr

  integer(kind=Int32)          :: ftfd0_hist(0:3,0:65535)  ! same FTCOARSE
  integer(kind=Int32)          :: ftfd1_hist(0:3,0:65535)  ! following FTCOARSE

  real(kind=8) :: p_quartz_1 = 20.48d-6, p_quartz_2 = 40.0d-9, ft(0:6), ff(0:6), ftav1(0:6)
  integer(kind=Int32) :: naux, clocks(0:6) = (/1834124,4979852,1191596,141795,149116,108612,0/)
  logical(kind=bool), parameter  :: ldebug = .TRUE. ! .FALSE.


  integer(kind=Int16), dimension(:,:), allocatable :: Ftcoarse 
  integer(kind=Int32), dimension(:,:), allocatable :: Ftfine
  integer(kind=Int32), dimension(:,:), allocatable :: Ftc_diff
  integer(kind=Int32), dimension(:,:), allocatable :: Ftf_diff

  integer(kind=Int16)          :: iccd
  integer(kind=Int32), dimension(:) :: ccdnr(1:12), nt(0:3)
  integer(kind=Int16), dimension(:), allocatable :: aux1ccdnr

  integer(kind=Int32), dimension(:), pointer :: aux1FrameCol => null()  
  integer(kind=Int16), dimension(:), pointer :: aux1FtcoarseCol => null()  
  integer(kind=Int16), dimension(:), pointer :: aux1FtfineCol => null()  
  integer(kind=Int8),  dimension(:), pointer :: aux1QuadrantCol => null()  
  integer(kind=Int8),  dimension(:), pointer :: aux1CCDidCol => null()  

  integer(kind=Int32), dimension(:), pointer :: aux2FrameCol => null()  
  integer(kind=Int16), dimension(:), pointer :: aux2NaboveCol => null()  
  integer(kind=Int16), dimension(:), pointer :: aux2NdefaCol => null()  

  type(DataSetT)               :: auxfile, ccxfile
  type(TableT)                 :: aux1table, aux2table, ccx1table

!---------------------------------------------------------------------------

  write(errstr,*) 'EPAUXCHECK ',trim(CodeVersion),' is running...'
  call message(AppMsg,Sparse)

  withexplanation    = booleanParameter("withexplanation")

  if (withexplanation) call printExplanation

  withpnauxname    = booleanParameter("withpnauxname")
  withpnauxname    = .TRUE.

  auxfileName      = trim(stringParameter("pnauxname"))

!---------------------------------------------------------------------------
! get name of AUXFILE from ODF, open file and check existence of extensions
! NB: access via ODF not yet implemented, only access via PNAUX filename 

  auxName = trim(auxfileName(index(auxfileName, '/', .true.)+1:))
  auxfile = dataSet(auxfileName,READ,USE_ENVIRONMENT)

  write(errstr,'(A)') '------------------------------------------------------------------------------------'
  call message(AppMsg,Sparse)
  write(errstr,*)' PNAUX file for check: ',trim(auxName)
  call message(AppMsg,Sparse)

!---------------------------------------------------------------------------
  hasPNAUX1 = hasBlock(auxfile,"PNAUX1")
  if (hasPNAUX1) then
    aux1table = table(auxfile,"PNAUX1")
    ntaux1 = numberOfRows(aux1table)
    write(errstr,*)' PNAUX1 has ',ntaux1,' number of rows'
    call message(AppMsg,Noisy)           ! number of rows (1..ntaux1)
  else                                   ! should never occur [who knows]
    ntaux1 = 0                           ! no rows
    call warning("noPNAUX1","AUX file does not contain extension PNAUX1.")
    write(errstr,*) 'The ODF [',ODF_icd,'] is invalid, please contact the originator of this data.'
    call warning('illegalODF')
    write(errstr,*) 'Several corrections cannot be applied in subsequent tasks. This may cause follow-up error messages.'
    call message(AppMsg,Sparse)
  endif

!---------------------------------------------------------------------------
  hasPNAUX2 = hasBlock(auxfile,"PNAUX2")
  if (hasPNAUX2) then
    aux2table = table(auxfile,"PNAUX2")
    ntaux2 = numberOfRows(aux2table)
    write(errstr,*)' PNAUX2 has ',ntaux2,' number of rows'
    call message(AppMsg,Noisy)           ! number of rows (1..ntaux2)
  else                                   ! should never occur [who knows]
    ntaux2 = 0                           ! no rows [unfortunately happens!]
    call warning("noPNAUX2","AUX file does not contain extension PNAUX2.")
    write(errstr,*) 'The ODF [',ODF_icd,'] is invalid, please contact the originator of this data.'
    call warning('illegalODF')
    write(errstr,*) 'Several corrections cannot be applied in subsequent tasks. This may cause follow-up error messages.'
    call message(AppMsg,Sparse)
  endif

!---------------------------------------------------------------------------
  ccdnr(1:12) = 0
  nt(0:3) = 0
  if (ntaux1 > 0) then

    allocate(aux1ccdnr(0:ntaux1-1))
    aux1table = table(auxfile,'PNAUX1')

    aux1FrameCol    =>  int32Data(column(aux1table,"CYCLE",READ))
    aux1FtcoarseCol =>  int16Data(column(aux1table,"FTCOARSE",READ))
    aux1FtfineCol   => uint16Data(column(aux1table,"FTFINE",READ))
    aux1QuadrantCol =>   int8Data(column(aux1table,"QUADRANT",READ))
    aux1CCDidCol    =>   int8Data(column(aux1table,"CCDID",READ))

    nffff = COUNT(MASK = aux1FtcoarseCol(0:ntaux1-1) == 32767)
    write(errstr,*)' PNAUX1 has ',nffff,' rows with FTCOARSE == 32767'
    call message(AppMsg,Verbose)

    aux1ccdnr(0:ntaux1-1) = aux1QuadrantCol(0:ntaux1-1)*3+aux1CCDidCol(0:ntaux1-1)+1
    do iccd = 1,12
      ccdnr(iccd) = COUNT(MASK = aux1ccdnr(0:ntaux1-1) == iccd)
      write(errstr,'(a,i3,a,i9,a)')'  PNAUX1 CCD',iccd,' has',ccdnr(iccd),' entries'
      call message(AppMsg,Verbose)
      n48828 = COUNT(MASK = aux1ccdnr(0:ntaux1-1) == iccd .AND. aux1Ftfinecol(0:ntaux1-1).gt.48828)
      write(errstr,*)'  PNAUX1 CCD',iccd,' has',n48828,' too large FTFINE entries'
      call message(AppMsg,Noisy)
      n32400 = COUNT(MASK = aux1ccdnr(0:ntaux1-1) == iccd .AND. &
                            aux1Ftcoarsecol(0:ntaux1-1).ge.32400 .and. &
                            aux1Ftcoarsecol(0:ntaux1-1).lt.32767)
      write(errstr,*)'  PNAUX1 CCD',iccd,' has',n32400,' too large FTCOARSE entries'
      call message(AppMsg,Noisy)
    enddo

    ntmax = 0
    do iquad = 0,3
      nt(iquad) = COUNT(MASK = (aux1QuadrantCol(0:ntaux1-1) == iquad) .AND. &
                               (aux1CCDidCol(0:ntaux1-1) == 0) .AND. &
                               (aux1FtcoarseCol(0:ntaux1-1) < 32767))
      write(errstr,'(A,i2,a,i9,a)')'  PNAUX1 QUAD',iquad,' has',nt(iquad),' non-FFFF entries'
      call message(AppMsg,Verbose)
      if (nt(iquad) .gt. ntmax) ntmax = nt(iquad)
    enddo


!-----------------------------------------------------------------------------------
!
! for the time being (and simplicity) only check quadrant 1 (with CCD 4):
!
!   we need  FTCOARSE and FTFINE
!   we need to check for FTCOARSE = 32767
!   we need to check for normal FTCOARSE reset
!
! FTFINE unit is 512 oscillator clocks 
!
! each EPIC-pn submode read-out sequence has a constant number of clocks, 
!   which is not an integer multiple of 512, therefore the difference in FTFINE 
!   between two consecutive events (of different read-out cycles) jitters
!
! However, the mean FTFINBE diffence value should be a constant throughout the
!   mission for events in the same FTCOARSE (i.e. independent from external clocks)
!
!-----------------------------------------------------------------------------------
!
! average FTFINE differences for each submode: this is FIX (due to readout sequence):
!   however, at changes of FTCOARSE (reset of FTFINE) different values can occur !!
!   this is due to 48828.125 FTFINE units per second (introducing additional jitter)
!   and (external) timing inaccuracies (plus clock frequency drifts).
! For reference, see https://xmmweb.esac.esa.int/docs/documents/CAL-TN-0081.pdf
!  (originally published as: Freyberg et al (2005), MPE Report Vol.288, 159-164)
!
!    3582.27344   FF         1834124
!   13822.27344  eFF_5       7077004 = 1834124 + 5*1048576
!    9726.27344  eFF_3       4979852 = 1834124 + 3*1048576
!    2327.33594   LW         1191596
!     276.94336   SW          141795
!     291.24219   TI          149116
!     212.13281   BU          108612
!
! NOTE: only event pairs across FTCOARSE boundaries can be used to estimate 
!   "effective" frequency drifts (which can vary even within an exposure due
!   due temperature drifts in the EPIC-pn quadrant box or external effects)
!
!-----------------------------------------------------------------------------------

    epn_submode(0) = 'FF' 
    epn_submode(1) = 'eFF_3' 
    epn_submode(2) = 'LW' 
    epn_submode(3) = 'SW' 
    epn_submode(4) = 'TI' 
    epn_submode(5) = 'BU' 
    epn_submode(6) = 'eFF_5'
    epn_submode(7) = 'ERROR' 

    allocate(ftcoarse(0:3,0:ntmax),ftfine(0:3,0:ntmax))
    allocate(ftc_diff(0:3,0:ntmax),ftf_diff(0:3,0:ntmax))

    ftcoarse = 0
    ftfine = 0
    ftc_max(0:3) = 0
    
    do iquad = 0,3
      it = 0
      ftcoarse(iquad,it) = nt(iquad)
      ftfine(iquad,it) = nt(iquad)
      do irow = 0,ntaux1-1
        if ((aux1QuadrantCol(irow) == iquad) .AND. &
                               (aux1CCDidCol(irow) == 0) .AND. &
                               (aux1FtcoarseCol(irow) < 32767)) then
          it = it+1
          ftcoarse(iquad,it) = aux1FtcoarseCol(irow)
          if (aux1FtfineCol(irow) .ge. 0) then
            ftfine(iquad,it) = aux1FtfineCol(irow)
          else ! negative
            ftfine(iquad,it) = aux1FtfineCol(irow)+65536 ! original value are unsigned I*2
          endif
!!!       if (ftcoarse(iquad,it) .ge. 32400) then
!!!         print*,irow,iquad,it,ftcoarse(iquad,it),ftfine(iquad,it),ftcoarse(iquad,it-1),ftfine(iquad,it-1)
!!!       endif
        endif
      enddo
!-----------
      ftc_max(iquad) = MAXVAL(ftcoarse(iquad,1:nt(iquad)))
      if (ftc_max(iquad) > 32399) then
        write(errstr,'(A,i2,a,i5,a,i4,a)') '  PNAUX1 QUAD',iquad,' has maximum FTCOARSE: ',ftc_max(iquad),&
              ' which could be due to time jumps of sum:', ftc_max(iquad)-32399,' [s]'
        call message(AppMsg,Sparse)   
      else
        write(errstr,'(A,i2,a,i5)') '  PNAUX1 QUAD',iquad,' has maximum FTCOARSE: ',ftc_max(iquad)
        call message(AppMsg,Verbose)   
      endif
!     print*,IQUAD,ftc_max(iquad)
!-----------
      if (it .ne. nt(iquad)) then
        write(errstr,*)'  PNAUX1 QUAD',iquad,' has',it,' good FTCOARSE entries but should have',nt(iquad)
        call message(AppMsg,Sparse)   
      endif
    enddo
    if (MINVAL(ftc_max) > 0) then
      if (MAXVAL(ftc_max) > MINVAL(ftc_max)) then
        write(errstr,*)'  PNAUX1 quadrants have different maximum FTCOARSE, maybe time jumps present.'
        call message(AppMsg,Sparse)
      endif
    endif
  write(errstr,'(A)') '------------------------------------------------------------------------------------'
  call message(AppMsg,Sparse)

!!! allocate(Time(1:naux),ftf_diff(1:naux),time_d(1:naux),time_i(1:naux),time_r(1:naux),ftfr_diff(1:naux))
!!! ftf_diff(1:naux-1) = FtFine(2:naux)-FtFine(1:naux-1)

    iquad = 1
    ftfd0_hist = 0
    ftfd1_hist = 0

    do iquad = 0,3
      do irow = 1, nt(iquad)-1
        fd = ftfine(iquad,irow+1)-ftfine(iquad,irow)
        if (FTcoarse(iquad,irow+1).eq.FTcoarse(iquad,irow)) then
          if (fd.gt.0) then
            ftfd0_hist(iquad,fd) = ftfd0_hist(iquad,fd) + 1
          endif
        elseif (FTcoarse(iquad,irow+1).eq.FTcoarse(iquad,irow)+1) then
!         print*,FTcoarse(iquad,irow),ftfine(iquad,irow),FTcoarse(iquad,irow+1),ftfine(iquad,irow+1)
          if (fd.lt.0) then
            ftfd1_hist(iquad,-fd) = ftfd1_hist(iquad,-fd) + 1
          endif
        endif
      enddo
    enddo

!------------------ this output should be written to a proper FITS file ---\
    if (ldebug) then                                                       !
                                                                           !
      open(unit=20,file='ftfd0.hist',form='formatted',status='unknown')    !
      do fd = 0, 65535                                                     !
        if (MAXVAL(ftfd0_hist(0:3,fd)).gt.0) then                          !
          write(20,'(1x,i5,4i10)') fd,(ftfd0_hist(iquad,fd),iquad=0,3)     !
        endif                                                              !
      enddo                                                                !
      close(20)                                                            !
                                                                           !
      open(unit=30,file='ftfd1.hist',form='formatted',status='unknown')    !
      do fd = 65535,0,-1                                                   !
        if (MAXVAL(ftfd1_hist(0:3,fd)).gt.0) then                          !
          write(30,'(1x,i5,4i10)') fd,(ftfd1_hist(iquad,fd),iquad=0,3)     !
        endif                                                              !
      enddo                                                                !
      close(30)                                                            !
                                                                           !
    endif                                                                  !
!------------------ that output should be written to a proper FITS file ---/

      clocks(6) = clocks(0)+5*((clocks(1)-clocks(0))/3)  ! eFF F1294=5
      ft(0:6) = clocks(0:6)*p_quartz_2*1000  ! nominal frame times [ms]
      ff(0:6) = clocks(0:6)/512.d0      ! nominal FTFINE differences
      ftfav = 0

!------------
! get instrument submode from average FTFINE difference
!
    do iquad = 0,3
      ih = 200               ! shortest mode (BU) has 212 FTFINE units
      hh1 = ftfd0_hist(iquad,ih-2)
      hh2 = ftfd0_hist(iquad,ih-1)
      do fd = ih,65535
        hh3 = ftfd0_hist(iquad,fd)
!       if (ldebug) then
!         if (fd.lt.10000) write(40,*)fd,hh1,hh2,hh3
!       endif
!
! the first 2 consecutive non-zero histogram values will be regarded as "proper"
!
        if ((hh3.gt.0.and.hh2.gt.0).or.(hh1.gt.0.and.hh2.gt.0)) then  ! at least 2 entries >0
          if (hh2.ge.hh1.and.hh2.ge.hh3) then
            ftfav(iquad) = (dble(fd-2)*hh1+dble(fd-1)*hh2+dble(fd)*hh3)/dble(hh1+hh2+hh3)
            exit
          endif
        elseif (hh3.eq.0.and.hh2.gt.0.and.hh1.eq.0) then              ! only 1 entriy >0
          write(errstr,'(A,I2,A,i6,A,3i9)') '  PNAUX1 QUAD',iquad,': unusual FTFINE difference histogram at DELTA(FT) =',&
              fd,' : frequencies +- 1FT:',hh1,hh2,hh3
          call message(AppMsg,Sparse)   
        endif
        hh1 = hh2
        hh2 = hh3
      enddo
!     print*,'Cycle end at',fd,ftfav(iquad)
      if (ftfav(iquad).gt.0) then
        write(errstr,'(A,I1,A,f13.5,A)') "Cycle time within same FTCOARSE: Q",iquad," :",ftfav(iquad)," FTFINE units"
        call message(AppMsg,Sparse)   
        write(errstr,'(A,I1,A,f13.6,A)') "Cycle time within same FTCOARSE: Q",iquad," :",ftfav(iquad)*20.48E-3," [ms] for 25 MHz"
        call message(AppMsg,Sparse)   
      endif
    enddo
!
! analysis above has been perfomred for all quadrants, but all (non-zero) values
! should agree within statistical uncertainties. Therefore try to match just the
! value for Q1 with nominal values and thus get the submode
!
    iquad = 1

      ldone = .false.
      do anchop = 0,50
        do ical = 0,6
          if (abs(ftfav(iquad)-ff(ical)*(1+anchop)).lt.1) then
            ldone = .true.
            exit
          endif
        enddo
        if (ldone) exit
      enddo

      if (.NOT. ldone) then
        write(errstr,'(A,I1,A,2i3,f13.6,A)') "Mode match not successful for Q",iquad," : ",ical,anchop,ftfav(iquad)," FTFINE units"
        call warning("ModeNotFound")
      endif

      if (anchop .eq. 0) then
        write(errstr,'(A,A5,A,f13.6,A)') &
             " Suggested mode:      ",epn_submode(ical)," nominal :",ft(ical)," [ms] for 25 MHz"
        call message(AppMsg,Sparse)   
      else
        write(errstr,'(A,A5,A,f13.6,A,I2)') &
             " Suggested mode:      ",epn_submode(ical)," nominal :",ft(ical)," [ms]  with Chopper = ",anchop
        call message(AppMsg,Sparse)   
      endif


!
! and now the FTFINE difference across FTCOARSE boundaries:
!   start at around the expected value "ftav1_exp" and compare with the measured value "ftav1_aux"
!
    ftav1_exp = 48828.125-ff(ical)*(1+anchop)  ! the larger the measured value the larger the real "frequency"
!   print*,'X:',ical,ftav1_exp
   
    iquad = 1
    do iquad = 0,3
      ih = int(ftav1_exp)-10
      hh1 = ftfd1_hist(iquad,ih-2)
      hh2 = ftfd1_hist(iquad,ih-1)
      do fd = ih,65535
        hh3 = ftfd1_hist(iquad,fd)
        if ((hh1.gt.0).and.(hh2.gt.0)) then 
!         print*,'A:',fd,hh1,hh2,hh3
!         if (hh3.gt.0) then
!           print*,'C:',fd,hh1,hh2,hh3
            ftfav(iquad) = 48828.125-(dble(fd-2)*hh1+dble(fd-1)*hh2+dble(fd)*hh3)/dble(hh1+hh2+hh3)
            exit
!         endif
        endif
        hh1 = hh2
        hh2 = hh3
      enddo
      if (ftfav(iquad).gt.0) then
        write(errstr,'(A,I1,A,f13.5,A)') "Cycle time across next FTCOARSE: Q",iquad," :",ftfav(iquad)," FTFINE units"
        call message(AppMsg,Sparse)   
        write(errstr,'(A,I1,A,f13.6,A)') "Cycle time across next FTCOARSE: Q",iquad," :",ftfav(iquad)*20.48E-3," [ms] for 25 MHz"
        call message(AppMsg,Sparse)
        freq_diff = (ff(ical)-ftfav(iquad)/(1+anchop))*512
        if (ical.le.6) then
          write(errstr,'(A,f12.6,A)')      "Effective oscillator frequency shift: ",freq_diff," [Hz]"
        else
          write(errstr,'(A,e17.6,A)')      "Effective oscillator frequency shift: ",freq_diff," [Hz]"
        endif
        call message(AppMsg,Sparse)
      endif
    enddo

    write(errstr,'(A)') '------------------------------------------------------------------------------------'
!!   call message(AppMsg,Sparse)

! 777  format(a19,2x,a4,1x,a10,1x,a6,1x,2i3,f15.7,f14.7,2i9,f16.7,f12.5,a4)
!      if (ical.gt.6.or.anchop.gt.0) then
!        write(40,777) dateobs,rev,obsid,expid,ical,anchop,ftfav
!        goto 996
!      endif


    iquad = 1
    do iquad = 0,3
        ftc_diff(iquad,1:nt(iquad)-1) = ftcoarse(iquad,2:nt(iquad))-ftcoarse(iquad,1:nt(iquad)-1)
        ftf_diff(iquad,1:nt(iquad)-1) =   ftfine(iquad,2:nt(iquad))-ftfine(iquad,1:nt(iquad)-1)
    enddo


!
! Negative time differences in full second counter naturally occur at
!   resets after FTCOARSE = 32399 s -> 0s to avoid I*2 overflow (15 bit).
! However, the value FTCOARSE = 32400 (or larger) should not occur, this
!   is an indication of (at least) one time jump in that last 32400s.
! Apparent negative time jumps of -1s occur if some data entries for a
!   quadrant appear twice in the list, then the negative jumps is very
!   close to a (negative) integer multiple of the cycle time (e.g. -14
!   cycle times in the case of FF mode). This is not a "classical" time
!   jump which would require a shift of FTCOARSE as corrective action,
!   but it is sufficient to igne the second occurence of the data in
!   PNAUX1 (and also in PNAUX2 !). Usually such occurences happen for
!   all active quadrants at the same time.
!
    iquad = 1
    do iquad = 0,3
      write(errstr,'(A)') '------------------------------------------------------------------------------------'
      if (nt(iquad).gt.1) call message(AppMsg,Sparse)
      do irow = 1,nt(iquad)-1
          ftd_frms = (ftc_diff(iquad,irow)*(48828.125)+ftf_diff(iquad,irow))/ff(ical)
          ftd_frac = ABS(ftd_frms-NINT(ftd_frms))
          ftd_frms = ((ftc_diff(iquad,irow)-1)*(48828.125)+ftf_diff(iquad,irow))/ff(ical)
          ftd_corr = ABS(ftd_frms-NINT(ftd_frms))
        if (ftc_diff(iquad,irow) < 0) then
          if  (ftc_diff(iquad,irow) < -32000) then
            write(errstr,'(A,i2,a,i9,A,2I7)') 'PNAUX1 QUAD',iquad,': at row',irow, &
                ' negative FTCOARSE/FTFINE difference:',ftc_diff(iquad,irow),ftf_diff(iquad,irow)
            call message(AppMsg,Sparse)
!           print*,iquad,irow,ftc_diff(iquad,irow),ftf_diff(iquad,irow)
          else
            write(errstr,'(A,i2,a,i9,A,2I7,a,2(i7,i6))') 'PNAUX1 QUAD',iquad,': at row',irow, &
                ' negative FTCOARSE/FTFINE difference:',ftc_diff(iquad,irow),ftf_diff(iquad,irow), &
                ' with FTC/FTF pairs before/after: ', &
                ftcoarse(iquad,irow),ftfine(iquad,irow),ftcoarse(iquad,irow+1),ftfine(iquad,irow+1)
            call message(AppMsg,Sparse)
!            print*,iquad,irow,ftc_diff(iquad,irow),ftf_diff(iquad,irow), &
!               ftcoarse(iquad,irow),ftfine(iquad,irow),ftcoarse(iquad,irow+1),ftfine(iquad,irow+1)
          endif
        elseif (ABS(ftd_frac).gt.0.05) then
            write(errstr,'(A,i2,a,i9,A,2I7,a,2(i7,i6))') 'PNAUX1 QUAD',iquad,': at row',irow, &
                ' positive FTCOARSE/FTFINE difference:',ftc_diff(iquad,irow),ftf_diff(iquad,irow), &
                ' with FTC/FTF pairs before/after: ', &
                ftcoarse(iquad,irow),ftfine(iquad,irow),ftcoarse(iquad,irow+1),ftfine(iquad,irow+1)
            call message(AppMsg,Sparse)

        endif
      enddo
    enddo

    write(errstr,'(A)') '------------------------------------------------------------------------------------'
    call message(AppMsg,Sparse)

!
! Forward time jumps are (mostly?) 1s and occur at the last read-out
!   cycle of a full second, i.e. occur when incrementing the FTCOARSE
!   counter (and resetting the FTFINE counter). So the time difference
!   between such two events is (mostly) 1s + 1 cycle time, e.g. 1.073s
!   for FF mode (or 14.63 cycle times).
! Such cases have a FTFINE value smaller than the cycle length (i.e.
!   the time word arrives soon after the reset, and another bit of the
!   FTCOARSE has incremented additionally (e.g., +32s, +64s, +128s).
! Counting mode intervals occur for each quadrant independently and
!   have a generic length for each submode (e.g. ~ 58s or 780 cycles
!   for FF mode).
! During counting mode also the time words related to scientific HK
!   values (every 20 cycles) stored in PNAUX2 are missing.
! Time jumps can occur also in counting mode intervals.
! Smaller gaps are mostly due to telemetry gaps (mostly due to ground
!   station issues) and thus occur for all quadrants simultaneously.
!   However, as the telemetry is transmitted quadrant-wwise in small
!   chunks of (order 1-3 s)  the actual gap times are then scattered 
!   by the chunk length and are not fully synchronuous.
!
!   print*,'----------------------------------------------------'
    iquad = 1
    do iquad = 0,3
      do irow = 1,nt(iquad)-1
        if (ftc_diff(iquad,irow) > 1) then
!!        print*,iquad,irow,ftc_diff(iquad,irow),ftf_diff(iquad,irow)
!          ftd_frms = (ftc_diff(iquad,irow)*(48828.125+ftfav(iquad)-ff(ical))+ftf_diff(iquad,irow))/ff(ical)
!          ftd_frac = ABS(ftd_frms-NINT(ftd_frms))
!          ftd_frms = ((ftc_diff(iquad,irow)-1)*(48828.125+ftfav(iquad)-ff(ical))+ftf_diff(iquad,irow))/ff(ical)
!          ftd_corr = ABS(ftd_frms-NINT(ftd_frms))
          ftd_frms = (ftc_diff(iquad,irow)*(48828.125)+ftf_diff(iquad,irow))/ff(ical)
          ftd_frac = ABS(ftd_frms-NINT(ftd_frms))
          ftd_frms = ((ftc_diff(iquad,irow)-1)*(48828.125)+ftf_diff(iquad,irow))/ff(ical)
          ftd_corr = ABS(ftd_frms-NINT(ftd_frms))
!          ftd_frms = (ftc_diff(iquad,irow)*48828.125+ftf_diff(iquad,irow))/ftfav(iquad)
!          ftd_frac = ABS(ftd_frms-NINT(ftd_frms))
!          ftd_frms = ((ftc_diff(iquad,irow)-1)*48828.125+ftf_diff(iquad,irow))/ftfav(iquad)
!          ftd_corr = ABS(ftd_frms-NINT(ftd_frms))

          if (ABS(ftd_frac).gt.0.05) then
!           write(errstr,'(A,i2,a,i9,A,2I7,a,F9.5,a,ES10.4)')  &
            write(errstr,'(A,i2,a,i9,A,2I7,a,F9.5,a,F9.5)')  &
              'PNAUX1 QUAD',iquad,': at row',irow, ' : Jump FTCOARSE/FTFINE ', &
            ftc_diff(iquad,irow),ftf_diff(iquad,irow), &
            ', fractional cycle time: ', ftd_frac,', after frequency correction: ',ftd_corr
            call message(AppMsg,Sparse)
!!!++ !           print*,iquad,irow,ftc_diff(iquad,irow),ftf_diff(iquad,irow),ftd_frac,ftd_corr
!!!++             write(errstr,'(A,i2,a,i9,A,2I7,a,2(i7,i6))') 'PNAUX1 QUAD',iquad,': at row',irow, &
!!!++                 ' positive FTCOARSE/FTFINE difference:',ftc_diff(iquad,irow),ftf_diff(iquad,irow), &
!!!++                 ' with FTC/FTF pairs before/after: ', &
!!!++                 ftcoarse(iquad,irow),ftfine(iquad,irow),ftcoarse(iquad,irow+1),ftfine(iquad,irow+1)
!!!++             call message(AppMsg,Sparse)
          endif


        endif
      enddo
    enddo








!    print*,'----------------------------------------------------'
   ! BOC2012/MK_AUX/mk_aux.f90

 996   continue

    if (allocated(ftcoarse)) deallocate (ftcoarse)
    if (allocated(ftfine))   deallocate (ftfine)
    if (allocated(ftc_diff)) deallocate (ftc_diff)
    if (allocated(ftf_diff)) deallocate (ftf_diff)

  endif

  if (allocated(aux1ccdnr)) deallocate(aux1ccdnr)



!---------------------------------------------------------
! check for potential buffer overflows in PNAUX2
!
  if (ntaux2 > 0) then
    write(errstr,'(A)') '------------------------------------------------------------------------------------'
    call message(AppMsg,Sparse)
    aux2table = table(auxfile,'PNAUX2')
    aux2FrameCol => int32Data(column(aux2table,"CYCLE",READ))
    aux2NaboveCol => uint16Data(column(aux2table,"NABOVE",READ))
    aux2NdefaCol => int16Data(column(aux2table,"NDEFA",READ))

    nloss = COUNT(MASK = aux2NaboveCol(0:ntaux2-1)-aux2NdefaCol(0:ntaux2-1) > 0)
    nover = COUNT(MASK = aux2NaboveCol(0:ntaux2-1)-aux2NdefaCol(0:ntaux2-1) > 511)

    write(errstr,*)' PNAUX2 has ',nloss,' rows with NDEFA < NABOVE'
    call message(AppMsg,Verbose)

    write(errstr,*)' PNAUX2 has ',nover,' rows with NABOVE - NDEFA > 511'
    call message(AppMsg,Verbose)
  endif


!---------------------------------------------------------
! release the (hopefully) opened PNAUX tables
!
  if (ntaux1 > 0) call release(aux1table)
  if (ntaux2 > 0) call release(aux2table)

  end subroutine  epauxcheck

  subroutine chk_pnaux
  end subroutine chk_pnaux



  subroutine printExplanation

    write(errstr,'(A)') "-------------------------------------------------------------------------------"
    call message(AppMsg,0)
    write(errstr,'(A)') " EPIC-pn has a master sequencer oscillator of nominal 25 MHz."
    call message(AppMsg,0)
    write(errstr,'(A)') " Time information is stored in the PNAUX1 extension in FTCOARSE FTFINE columns."
    call message(AppMsg,0)
    write(errstr,'(A)') " The unit of sub-second counter FTFINE is 512 oscillator clocks [20.48 mu-s]."
    call message(AppMsg,0)
    write(errstr,'(A)') " Each submode lasts a fixed number of oscillator clocks."
    call message(AppMsg,0)
    write(errstr,'(A)') " These numbers are not integer multiples of 512 clocks and therefore the"
    call message(AppMsg,0)
    write(errstr,'(A)') " integer FTFINE values of the event time stamps show a jitter of +-1 unit."
    call message(AppMsg,0)
    write(errstr,'(A)') " Moreover, also a full second is not an integer number of FTFINE units."
    call message(AppMsg,0)
    write(errstr,'(A)') " For a nominal frequency of 25 MHz there are 48828.125 FTFINE units per second."
    call message(AppMsg,0)
    write(errstr,'(A)') " This causes another jitter for FTFINE differences for consecutive events"
    call message(AppMsg,0)
    write(errstr,'(A)') " across FTCOARSE boundaries."
    call message(AppMsg,0)
    write(errstr,'(A)') " The submode values are as follows:"
    call message(AppMsg,0)
    write(errstr,'(A)') "    SubMode     Clocks       FTFINE        Nominal 25 MHz"
    call message(AppMsg,0)
    write(errstr,'(A)') "      FF       1834124      3582.27344      73.36496 [ms]"
    call message(AppMsg,0)
    write(errstr,'(A)') "     eFF_5     7077004     13822.27344     283.08016 [ms]"
    call message(AppMsg,0)
    write(errstr,'(A)') "     eFF_3     4979852      9726.27344     199.19408 [ms]"
    call message(AppMsg,0)
    write(errstr,'(A)') "      LW       1191596      2327.33594      47.66384 [ms]"
    call message(AppMsg,0)
    write(errstr,'(A)') "      SW        141795       276.94336       5.67180 [ms]"
    call message(AppMsg,0)
    write(errstr,'(A)') "      TI        149116       291.24219       5.96464 [ms]"
    call message(AppMsg,0)
    write(errstr,'(A)') "      BU        108612       212.13281       4.34448 [ms]"
    call message(AppMsg,0)
    write(errstr,'(A)') " For consecutive events within the same FTCOARSE above values should apply."
    call message(AppMsg,0)
    write(errstr,'(A)') " The EPIC-pn full second counter FTCOARSE sometimes shows jumps of eiterh sign."
    call message(AppMsg,0)
    write(errstr,'(A)') " To correct for these one has to check for time differences that are non-integer"
    call message(AppMsg,0)
    write(errstr,'(A)') " multiples of the frame time. For long gaps this requires accurate knowledge of" 
    call message(AppMsg,0)
    write(errstr,'(A)') " the actual frame time, otherwise real time jumps escape detection or false ones"
    call message(AppMsg,0)
    write(errstr,'(A)') " are introduced. Typical gap sizes within a quadrant are of oorder <60s, while"  
    call message(AppMsg,0)
    write(errstr,'(A)') " telemetry gaps can last for ks, i.e. thousands of individual frame times."
    call message(AppMsg,0)
    write(errstr,'(A)') " For details see: https://xmmweb.esac.esa.int/docs/documents/CAL-TN-0081.pdf"
    call message(AppMsg,0)
    write(errstr,'(A)') "-------------------------------------------------------------------------------"
    call message(AppMsg,0)

  end subroutine printExplanation


end module epauxcheck_mod
